const slack = require('../slack')

const tie = 'TIE'

const messageFromBoard = ({ dataName, gameName, textFromBoard, board, player1, player2 }) =>
    gameName + ' between ' + player1.toUpperCase() + ' and ' + player2.toUpperCase() + ' ' + encodeGame(dataName, board, [player1, player2]) + '\n' +
    '```' + textFromBoard(board) + '\n```'

const addChoiceEmojis = async ({ choices, channel, ts }) => {
  const addEmoji = async emojiName =>
      await slack.app.client.reactions.add({
        channel,
        timestamp: ts,
        name: emojiName
      })
  for (const choice of choices) {
    await addEmoji(choice)
  }
}

const buildGameStarter = ({ startTriggers, dataName, gameName, textFromBoard, initialBoard, turnChoiceEmojis }) => async ({ event, say }) => {
  if (event.channel_type === 'im') {
    const eventText = event.text?.toLowerCase()
    if (eventText && startTriggers.find(keyword => eventText.startsWith('!' + keyword))) {
      const opponent = event.text.toUpperCase().match(/<@[^>]*>/)[0]
      const msg = messageFromBoard({
        dataName,
        gameName,
        textFromBoard,
        board: initialBoard(),
        player1: '<@' + event.user + '>',
        player2: opponent
      })
      const sent = await say(msg)
      await addChoiceEmojis({...sent, choices: turnChoiceEmojis})
    }
  }
}

const encodeGame = (dataKey, board, players) => slack.encodeData(dataKey, {board, players})

const decodeGame = (dataKey, message) => slack.decodeData(dataKey, message)

const getMessages = winner => {
  const buildMessage = state => !winner ? '' : winner === tie ? '\nIt\'s a tie!' : `\nYou ${state}!`
  return {
    you: buildMessage('win'),
    opponent: buildMessage('lost')
  }
}

const buildTurnHandler = ({ gameName, dataName, checkWinner, textFromBoard, turnChoiceEmojis, makeMove }) => async ({ event, say }) => {
  if (event.item_user !== slack.hvackerBotUserId || !turnChoiceEmojis.includes(event.reaction)) {
    return
  }

  const message = await slack.getMessage(event.item)
  if (!message.messages[0]?.text?.includes(gameName)) {
    return
  }

  const game = decodeGame(dataName, message.messages[0].text)
  if (!game) {
    return
  }

  const { board, players } = game
  let winner = checkWinner(board)
  if (winner) {
    return
  }

  const [player1, player2] = players
  let opponent = (player1.includes(event.user) ? player2 : player1)
  opponent = opponent.replace(/[^A-Z0-9]/gi, '').toUpperCase()

  if (!makeMove(event.reaction, board)) {
    await say('You can\'t go there!')
    return
  }
  winner = checkWinner(board)

  const boardMessage = messageFromBoard({
    dataName,
    gameName,
    textFromBoard,
    board,
    player1,
    player2
  })
  const winnerMessages = getMessages(winner)
  await say(boardMessage + winnerMessages.you)
  if (!winner) {
    await say('Waiting for opponent\'s response...')
  }

  const removeEmoji = async emojiName =>
      slack.app.client.reactions.remove({
        channel: event.item.channel,
        timestamp: message.messages[0]?.ts,
        name: emojiName
      })
  turnChoiceEmojis.forEach(removeEmoji)
  const sentBoard = await slack.app.client.chat.postMessage({
    channel: opponent,
    text: boardMessage + winnerMessages.opponent
  })
  if (!winner) {
    await addChoiceEmojis({...sentBoard, choices: turnChoiceEmojis})
  }
}

module.exports = {
  tie,
  build: ({
            startTriggers,
            initialBoard,
            turnChoiceEmojis,
            gameName,
            textFromBoard,
            checkWinner,
            makeMove,
          }) => {
    const dataName = gameName.replace(/[^0-9a-zA-Z]/gi, '')
    const gameStarter = buildGameStarter({
      startTriggers,
      gameName,
      dataName,
      initialBoard,
      textFromBoard,
      turnChoiceEmojis
    });
    slack.onMessage(gameStarter)

    const turnHandler = buildTurnHandler({
      gameName,
      dataName,
      checkWinner,
      textFromBoard,
      turnChoiceEmojis,
      makeMove
    })
    slack.onReaction(turnHandler)
  }
}
